/****************************************************************************
* VCGLib                                                            o o     *
* Visual and Computer Graphics Library                            o     o   *
*                                                                _   O  _   *
* Copyright(C) 2004-2009                                           \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/

#include<vcg/simplex/vertex/base.h>
#include<vcg/simplex/vertex/component.h>

#include <vcg/complex/used_types.h>

#include<vcg/simplex/face/base.h>
#include<vcg/simplex/face/component.h>

#include<vcg/simplex/face/topology.h>
#include<vcg/complex/trimesh/base.h>

// input output
#include<wrap/io_trimesh/import.h>
#include<wrap/io_trimesh/export.h>

// topology computation
#include<vcg/complex/trimesh/update/topology.h>

// normals
#include<vcg/complex/trimesh/update/normal.h>

using namespace vcg;
using namespace std;

class MyEdge;
class MyFace;
class MyVertex;
struct MyUsedTypes : public UsedTypes<	Use<MyVertex>   ::AsVertexType,
                                        Use<MyEdge>     ::AsEdgeType,
                                        Use<MyFace>     ::AsFaceType>{};

class MyVertex  : public Vertex<MyUsedTypes, vertex::Coord3f, vertex::Normal3f, vertex::BitFlags  >{};
class MyFace    : public Face< MyUsedTypes, face::FFAdj,  face::VertexRef, face::BitFlags > {};
class MyEdge    : public Edge<MyUsedTypes>{};
class MyMesh    : public tri::TriMesh< vector<MyVertex>, vector<MyFace> , vector<MyEdge>  > {};


// Return true if fp is the face with minimal index (pointer)
// wrt to the ring of faces incident on the edge <edgeInd>

bool IsMinimalInFaceRing(MyMesh::FacePointer fp, int edgeInd)
{
  MyMesh::FacePointer fpStart = fp;
  MyMesh::FacePointer fpCur = fp;
  int eiCur=edgeInd;
  do
  {
    MyMesh::FacePointer fpNext=fpCur->FFp(eiCur);
    int eiNext=fpCur->FFi(eiCur);
    if(fp>fpNext) return false;
    eiCur=eiNext;
    fpCur=fpNext;
  }
  while(fpCur!=fpStart);
  return true;
}

int CountIncidentFaces(MyMesh::FacePointer fp, int edgeInd)
{
  MyMesh::FacePointer fpStart = fp;
  MyMesh::FacePointer fpCur = fp;
  int eiCur=edgeInd;
  int incidentCnt=0;
  do
  {
    MyMesh::FacePointer fpNext=fpCur->FFp(eiCur);
    int eiNext=fpCur->FFi(eiCur);
    incidentCnt++;
    eiCur=eiNext;
    fpCur=fpNext;
  }
  while(fpCur!=fpStart);

  return incidentCnt;
}

void DetachSingleFace(MyMesh::FacePointer fp, int edgeInd)
{
  if(fp->FFp(edgeInd)==fp) return;

  MyMesh::FacePointer fpStart = fp;
  MyMesh::FacePointer fpCur = fp;
  int eiCur=edgeInd;
  do
  {
    MyMesh::FacePointer fpNext=fpCur->FFp(eiCur);
    int eiNext=fpCur->FFi(eiCur);
    eiCur=eiNext;
    fpCur=fpNext;
  }
  while(fpCur->FFp(eiCur)!=fpStart);

  MyMesh::FacePointer fpFirst = fp->FFp(edgeInd);
                  int eiFirst = fp->FFi(edgeInd);

  fp->FFp(edgeInd) = fp;
  fp->FFi(edgeInd) = edgeInd;

  fpCur->FFp(eiCur) = fpFirst;
  fpCur->FFi(eiCur) = eiFirst;
}

void DetachAllIncidentFace(MyMesh::FacePointer fp, int edgeInd)
{
  if(fp->FFp(edgeInd)==fp) return;
  MyMesh::FacePointer fpStart = fp;
  MyMesh::FacePointer fpCur = fp->FFp(edgeInd);
  int eiCur=fp->FFi(edgeInd);
  do
  {
    MyMesh::FacePointer fpNext=fpCur->FFp(eiCur);
    int eiNext=fpCur->FFi(eiCur);
    DetachSingleFace(fpCur,eiCur);
    eiCur=eiNext;
    fpCur=fpNext;
  }
  while(fpCur->FFp(eiCur)!=fpStart);
}

int main( int argc, char **argv )
{
  if(argc<2)
  {
    printf("Usage trimesh_base <meshfilename.ply>\n");
    return -1;
  }

  MyMesh m;

  if(tri::io::ImporterPLY<MyMesh>::Open(m,argv[1])!=0)
  {
    printf("Error reading file  %s\n",argv[1]);
    exit(0);
  }

  tri::UpdateTopology<MyMesh>::FaceFace(m);
  tri::UpdateFlags<MyMesh>::FaceClearV(m);
  tri::UpdateNormals<MyMesh>::PerVertexNormalized(m);
  int k = m.vn - ( m.fn*3/2) + m.fn;


  double nonManifNum=0;
  int nonManifNum2=0;
  int nonManifNum3=0;

  MyMesh::FaceIterator fi;
  for(fi=m.face.begin();fi!=m.face.end();++fi)
  {
    for(int i=0;i<3;++i)
    {
      // To set a face as visited
      // (*fi).SetV();
      // To test a face if it is visited
      // (*fi).IsV();
      // (*fi).ClearV();
      int nonManifDegree = CountIncidentFaces(&*fi,i);
      if(nonManifDegree>2)
      {
        nonManifNum += 1.0 / double(nonManifDegree);
        DetachAllIncidentFace(&*fi,i);
        if(IsMinimalInFaceRing(&*fi,i)) nonManifNum2++;
        nonManifNum3++;
      }
    }
  }


  printf("Input mesh  vn:%i en: %i fn:%i  k = %i \n",m.vn, m.fn*3/2, m.fn,k);


  printf( "Mesh has %lf non manifold edges\n", nonManifNum );
  printf( "Mesh has %i non manifold edges\n", nonManifNum2 );
  printf( "Mesh has %i non manifold edges\n", nonManifNum3 );

  /* Todo
     1) Modificare in modo da staccare solo le facce strettamente necessarie.
     2) Scegliere quale faccia staccare in modo consapevole:
       - quella che ha il maggior numero di edge non manifold o bordi
       - quella di area minima
     */


  return 0;
}
